@@ -1,5 +1,5 @@
 /* Handle so called `shell archives'.
-   Copyright (C) 1994, 1995 Free Software Foundation, Inc.
+   Copyright (C) 1994, 1995, 1996, 2002 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,10 +13,33 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
+#include "liballoca.h"
+
 #include "system.h"
+#include "basename.h"
+#include "error.h"
+#include "exit.h"
+#include "xalloc.h"
+#include "xgetcwd.h"
+
+#if HAVE_LOCALE_H
+# include <locale.h>
+#else
+# define setlocale(Category, Locale)
+#endif
+#include "gettext.h"
+#define _(str) gettext (str)
+#define N_(str) gettext_noop (str)
+
+#ifndef NUL
+#  define NUL '\0'
+#endif
 
 static const char *cut_mark_line
   = "---- Cut Here and feed the following to sh ----\n";
@@ -34,7 +57,7 @@ static const char *cut_mark_line
 /* Shell command able to count characters from its standard input.  We
    have to take care for the locale setting because wc in multi-byte
    character environments get different results.  */
-#define CHARACTER_COUNT_COMMAND "LC_ALL= LC_CTYPE= LANG= wc -c <"
+#define CHARACTER_COUNT_COMMAND "LC_ALL=C wc -c <"
 
 /* Command computing MD5 sumof specified file.  Because there is not
    official standard for this we have to restrict ourself to a few
@@ -67,6 +90,27 @@ static const char *cut_mark_line
 
 struct tm *localtime ();
 
+#ifdef HAVE_LIMITS_H
+# include <limits.h>
+#endif
+
+/* Determine whether an integer type is signed, and its bounds.
+   This code assumes two's (or one's!) complement with no holes.  */
+
+/* The extra casts work around common compiler bugs,
+   e.g. Cray C 5.0.3.0 when t == time_t.  */
+#ifndef TYPE_SIGNED
+# define TYPE_SIGNED(t) (! ((t) 0 < (t) -1))
+#endif
+#ifndef TYPE_MINIMUM
+# define TYPE_MINIMUM(t) ((t) (TYPE_SIGNED (t) \
+			       ? ~ (t) 0 << (sizeof (t) * CHAR_BIT - 1) \
+			       : (t) 0))
+#endif
+#ifndef TYPE_MAXIMUM
+# define TYPE_MAXIMUM(t) ((t) (~ (t) 0 - TYPE_MINIMUM (t)))
+#endif
+
 #if !NO_WALKTREE
 
 /* Declare directory reading routines and structures.  */
@@ -101,6 +145,7 @@ struct dirent *readdir ();
 /* Option variables.  */
 
 #include "getopt.h"
+#include "inttostr.h"
 #include "md5.h"
 
 /* No Brown-Shirt mode.  */
@@ -118,6 +163,9 @@ static int gzipped_file_mode = -1;
 /* -N option to gzip.  */
 static int gzip_compression_level = 9;
 
+/* Run input files through bzip2 (requires uuencoded_file_mode).  */
+static int bzipped_file_mode = -1;
+
 /* Run input files through compress (requires uuencoded_file_mode).  */
 static int compressed_file_mode = -1;
 
@@ -188,8 +236,8 @@ static int debugging_mode = 0;
 /* Split files in the middle.  */
 static int split_file_mode = 0;
 
-/* File size limit in kilobytes.  */
-static unsigned file_size_limit = 0;
+/* File size limit in bytes.  */
+static off_t file_size_limit = 0;
 
 /* Other global variables.  */
 
@@ -209,13 +257,13 @@ static FILE *output = NULL;
 static off_t archive_type_position;
 
 /* Position for first file in the shar file.  */
-static long first_file_position;
+static off_t first_file_position;
 
 /* Base for output filename.  */
-static char *output_base_name;
+static char *output_base_name = NULL;
 
 /* Actual output filename.  */
-static char *output_filename;
+static char *output_filename = NULL;
 
 static char *submitter_address = NULL;
 
@@ -236,14 +284,22 @@ static int sharpid;
 
 #if DEBUG
 # define DEBUG_PRINT(Format, Value) \
-    if (debugging_mode) printf(Format, Value)
+    if (debugging_mode)					\
+      {							\
+	char buf[INT_BUFSIZE_BOUND (off_t)];		\
+	printf (Format, offtostr (Value, buf));		\
+      }
 #else
 # define DEBUG_PRINT(Format, Value)
 #endif
 
 static void open_output __P ((void));
 static void close_output __P ((void));
-static void usage __P ((int));
+static void usage __P ((int))
+#if defined __GNUC__ && ((__GNUC__ == 2 && __GNUC_MINOR__ >= 5) || __GNUC__ > 2)
+     __attribute__ ((noreturn))
+#endif
+;
 
 /* Walking tree routines.  */
 
@@ -328,9 +384,11 @@ walkdown (routine, local_name, restore_name)
 
 	/* Avoid restoring `./xxx' when shar'ing `.'.  */
 
-	if (!strncmp (restore_name, "./", 2))
+	if (!strncmp (restore_name_copy, "./", 2))
 	  {
-	    strcpy (restore_name_copy, restore_name_copy + 2);
+	    const char *rp = restore_name_copy + 2;
+	    char *wp = restore_name_copy;
+	    while ((*wp++ = *rp++));
 	    restore_name_length -= 2;
 	  }
 
@@ -382,7 +440,7 @@ walktree (routine, local_name)
   memcpy (local_name_copy, local_name, len + 1);
   cursor = local_name_copy + len - 1;
   while (*cursor == '/' && cursor > local_name_copy)
-    *cursor-- = '\0';
+    *cursor-- = NUL;
 
   /* Remove crumb at beginning.  */
 
@@ -477,11 +535,9 @@ do\n\
   if test \"$gettext_dir\" = FAILED && test -f $dir/gettext \\\n\
      && ($dir/gettext --version >/dev/null 2>&1)\n\
   then\n\
-    set `$dir/gettext --version 2>&1`\n\
-    if test \"$3\" = GNU\n\
-    then\n\
-      gettext_dir=$dir\n\
-    fi\n\
+    case `$dir/gettext --version 2>&1 | sed 1q` in\n\
+      *GNU*) gettext_dir=$dir ;;\n\
+    esac\n\
   fi\n\
   if test \"$locale_dir\" = FAILED && test -f $dir/%s \\\n\
      && ($dir/%s --print-text-domain-dir >/dev/null 2>&1)\n\
@@ -505,31 +561,32 @@ fi\n",
 	 --print-text-domain-dir option has to be given.  */
     }
 
-  if (query_user_mode)
-    if (vanilla_operation_mode)
-      fputs ("\
+  if (!quiet_unshar_mode)
+    {
+      if (vanilla_operation_mode)
+	fputs ("\
 shar_tty= shar_n= shar_c='\n\
 '\n",
-	     output);
-    else
-      {
+	       output);
+      else
+	{
+	  if (query_user_mode)
+	    /* Check if /dev/tty exists.  If yes, define shar_tty to
+	       `/dev/tty', else, leave it empty.  */
 
-	/* Check if /dev/tty exists.  If yes, define shar_tty to
-	   `/dev/tty', else, leave it empty.  */
-
-	fputs ("\
+	    fputs ("\
 if test -n \"`ls /dev/tty 2>/dev/null`\"; then\n\
   shar_tty=/dev/tty\n\
 else\n\
   shar_tty=\n\
 fi\n",
-	       output);
+		   output);
 
-	/* Try to find a way to echo a message without newline.  Set
-	   shar_n to `-n' or nothing for an echo option, and shar_c to
-	   `\c' or nothing for a string terminator.  */
+	  /* Try to find a way to echo a message without newline.  Set
+	     shar_n to `-n' or nothing for an echo option, and shar_c
+	     to `\c' or nothing for a string terminator.  */
 
-	fputs ("\
+	  fputs ("\
 if (echo \"testing\\c\"; echo 1,2,3) | grep c >/dev/null; then\n\
   if (echo -n testing; echo 1,2,3) | sed s/-n/xn/ | grep xn >/dev/null; then\n\
     shar_n= shar_c='\n\
@@ -540,8 +597,9 @@ if (echo \"testing\\c\"; echo 1,2,3) | grep c >/dev/null; then\n\
 else\n\
   shar_n= shar_c='\\c'\n\
 fi\n",
-	       output);
-      }
+		 output);
+	}
+    }
 
   if (timestamp_mode)
     {
@@ -556,11 +614,11 @@ fi\n",
 	 Otherwise, set shar_touch to `:' and issue a warning.  */
 
       fprintf (output, "\
-if touch -am -t %s %s >/dev/null 2>&1 && test ! -f %s -a -f %s; then\n\
+if touch -am -t %s %s >/dev/null 2>&1 && test ! -f %s && test -f %s; then\n\
   shar_touch='touch -am -t $1$2$3$4$5$6.$7 \"$8\"'\n\
-elif touch -am %s %s >/dev/null 2>&1 && test ! -f %s -a ! -f %s -a -f %s; then\n\
+elif touch -am %s %s >/dev/null 2>&1 && test ! -f %s && test ! -f %s && test -f %s; then\n\
   shar_touch='touch -am $3$4$5$6$1$2.$7 \"$8\"'\n\
-elif touch -am %s %s >/dev/null 2>&1 && test ! -f %s -a -f %s; then\n\
+elif touch -am %s %s >/dev/null 2>&1 && test ! -f %s && test -f %s; then\n\
   shar_touch='touch -am $3$4$5$6$2 \"$8\"'\n\
 else\n\
   shar_touch=:\n\
@@ -581,61 +639,105 @@ installing GNU \\`touch', distributed in GNU File Utilities..."),
 	       stamp1, stamp2, stamp2tr, stamp3, file);
     }
 
-  if (!split_file_mode || part_number == 1)
+  if (file_size_limit == 0 || part_number == 1)
     {
       /* Create locking directory.  */
-      fprintf (output, "\
+      if (vanilla_operation_mode)
+	fprintf (output, "\
+mkdir _sh%05d || ( echo '%s' '_sh%05d'; exit 1 )\n",
+		 sharpid, N_("failed to create locking directory"), sharpid);
+      else
+	fprintf (output, "\
+$echo $shar_n 'x -' '%s' \"\\`_sh%05d': \"$shar_c\n\
 if mkdir _sh%05d; then\n\
-  $echo 'x -' '%s'\n\
+  $echo '%s'\n\
 else\n\
   $echo '%s'\n\
   exit 1\n\
 fi\n",
-	       sharpid, N_("creating lock directory"),
-	       N_("failed to create lock directory"));
+		 N_("lock directory"), sharpid, sharpid, N_("created"),
+		 N_("failed to create"));
     }
 }
 
-/*---.
-| ?  |
-`---*/
+/*----------------------------------------------.
+| generate_mkdir                                |
+| Make sure it is done only once for each dir   |
+`----------------------------------------------*/
 
-/* Ridiculously enough.  FIXME: No fix limit in GNU...  */
-#define MAX_MKDIR_ALREADY	128
-
-char *mkdir_already[MAX_MKDIR_ALREADY];
-int mkdir_already_count = 0;
+static int    mkdir_alloc_ct = 0;
+static int    mkdir_already_ct = 0;
+static char** mkdir_already;
 
 static void
 generate_mkdir (path)
      const char *path;
 {
-  int counter;
-  char *cursor;
-
   /* If already generated code for this dir creation, don't do again.  */
 
-  for (counter = 0; counter < mkdir_already_count; counter++)
-    if (!strcmp (path, mkdir_already[counter]))
-      return;
+  {
+    int    ct = mkdir_already_ct;
+    char** pp = mkdir_already;
+
+    while (--ct > 0)
+      {
+        if (strcmp (*(pp++), path) == 0)
+          return;
+      }
+  }
 
   /* Haven't done this one.  */
 
-  if (mkdir_already_count == MAX_MKDIR_ALREADY)
-    error (EXIT_FAILURE, 0, _("Too many directories for mkdir generation"));
-  cursor = mkdir_already[mkdir_already_count++] = xmalloc (strlen (path) + 1);
-  strcpy (cursor, path);
+  if (++mkdir_already_ct > mkdir_alloc_ct)
+    {
+      /*
+       *  We need more name space.  Get larger and larger chunks of space.
+       *  The bound is when integers go negative.  Too many directories.  :)
+       *
+       *  16, 40, 76, 130, 211, 332, 514, 787, 1196, 1810, 2731, 4112, ...
+       */
+      mkdir_alloc_ct += 16 + (mkdir_alloc_ct/2);
+      if (mkdir_alloc_ct < 0)
+        error (EXIT_FAILURE, 0, _("Too many directories for mkdir generation"));
+
+      if (mkdir_already != NULL)
+        mkdir_already =
+          xrealloc (mkdir_already, mkdir_alloc_ct * sizeof (char*));
+      else
+        mkdir_already = xmalloc (mkdir_alloc_ct * sizeof (char*));
+    }
+
+  /* Add the directory into our "we've done this already" table */
+
+  mkdir_already[ mkdir_already_ct-1 ] = xstrdup (path);
 
   /* Generate the text.  */
 
   fprintf (output, "if test ! -d '%s'; then\n", path);
   if (!quiet_unshar_mode)
-    fprintf (output, "  $echo 'x -' '%s' '%s'\n",
-	     N_("creating directory"), path);
-  fprintf (output, "  mkdir '%s'\n", path);
+    fprintf (output, "\
+  $echo $echo_n 'x -' '%s: '$echo_c\n\
+  if mkdir '%s'; then $echo '%s'; else $echo '%s'; fi\n",
+             path, path, N_("created"), N_("failed to create"));
+  else
+    fprintf (output, "  mkdir '%s'\n", path);
   fputs ("fi\n", output);
 }
 
+static void
+clear_mkdir_already (void)
+{
+  char** pp = mkdir_already;
+  int    ct = mkdir_already_ct;
+
+  mkdir_already_ct = 0;
+  while (--ct >= 0)
+    {
+      free (*pp);
+      *(pp++) = NULL;
+    }
+}
+
 /*---.
 | ?  |
 `---*/
@@ -696,7 +798,8 @@ generate_one_header_line (local_name, restore_name)
      const char *local_name;
      const char *restore_name;
 {
-  fprintf (output, "# %6ld %s %s\n", struct_stat.st_size,
+  char buf[INT_BUFSIZE_BOUND (off_t)];
+  fprintf (output, "# %6s %s %s\n", offtostr (struct_stat.st_size, buf),
 	   mode_string (struct_stat.st_mode), restore_name);
   return 0;
 }
@@ -828,7 +931,7 @@ generate_full_header (argc, argv)
       /* May be split, explain.  */
 
       fputs ("#\n", output);
-      archive_type_position = ftell (output);
+      archive_type_position = ftello (output);
       fprintf (output, "%-75s\n%-75s\n", "#", "#");
     }
 
@@ -891,8 +994,7 @@ shar (local_name, restore_name)
 {
   char buffer[BUFSIZ];
   FILE *input;
-  long current_size;
-  long remaining_size;
+  off_t remaining_size;
   int split_flag = 0;		/* file split flag */
   const char *file_type;	/* text of binary */
   const char *file_type_remote;	/* text or binary, avoiding locale */
@@ -915,9 +1017,9 @@ shar (local_name, restore_name)
 
   if (file_size_limit)
     {
-      current_size = ftell (output);
-      remaining_size = (file_size_limit * 1024L) - current_size;
-      DEBUG_PRINT (_("In shar: remaining size %ld\n"), remaining_size);
+      off_t current_size = ftello (output);
+      remaining_size = file_size_limit - current_size - 1024;
+      DEBUG_PRINT (_("In shar: remaining size %s\n"), remaining_size);
 
       if (!split_file_mode && current_size > first_file_position
 	  && ((uuencoded_file_mode
@@ -928,8 +1030,8 @@ shar (local_name, restore_name)
 
 	  /* Change to another file.  */
 
-	  DEBUG_PRINT (_("Newfile, remaining %ld, "), remaining_size);
-	  DEBUG_PRINT (_("Limit still %d\n"), file_size_limit);
+	  DEBUG_PRINT (_("New file, remaining %s, "), remaining_size);
+	  DEBUG_PRINT (_("Limit still %s\n"), file_size_limit);
 
 	  /* Close the "&&" and report an error if any of the above
 	     failed.  */
@@ -953,8 +1055,7 @@ exit 0\n",
 
 	  /* Clear mkdir_already in case the user unshars out of order.  */
 
-	  while (mkdir_already_count > 0)
-	    free (mkdir_already[--mkdir_already_count]);
+	  clear_mkdir_already ();
 
 	  /* Form the next filename.  */
 
@@ -981,7 +1082,7 @@ exit 0\n",
 
 	  generate_configure ();
 
-	  first_file_position = ftell (output);
+	  first_file_position = ftello (output);
 	}
     }
   else
@@ -1058,10 +1159,13 @@ exit 0\n",
 	  static int pid, pipex[2];
 
 	  file_type = (compressed_file_mode ? _("compressed")
-		       : gzipped_file_mode ? _("gzipped") : _("binary"));
+		       : gzipped_file_mode ? _("gzipped")
+		       : bzipped_file_mode ? _("bzipped")
+		       : _("binary"));
 	  file_type_remote = (compressed_file_mode ? N_("(compressed)")
 			      : gzipped_file_mode ? N_("(gzipped)")
-			        : N_("(binary)"));
+			      : bzipped_file_mode ? N_("(bzipped)")
+			      : N_("(binary)"));
 
 	  /* Fork a uuencode process.  */
 
@@ -1102,6 +1206,11 @@ exit 0\n",
 			   gzip_compression_level, local_name);
 		  input = popen (buffer, "r");
 		}
+	      else if (bzipped_file_mode)
+		{
+		  sprintf (buffer, "bzip2 < '%s'", local_name);
+		  input = popen (buffer, "r");
+		}
 	      else
 		input = fopen (local_name, "rb");
 
@@ -1111,11 +1220,14 @@ exit 0\n",
 		fprintf (outptr, "_sh%05d/cmp\n", sharpid);
 	      else if (gzipped_file_mode)
 		fprintf (outptr, "_sh%05d/gzi\n", sharpid);
+	      else if (bzipped_file_mode)
+		fprintf (outptr, "_sh%05d/bzi\n", sharpid);
 	      else
 		fprintf (outptr, "%s\n", restore_name);
 	      copy_file_encoded (input, outptr);
 	      fprintf (outptr, "end\n");
-	      if (compressed_file_mode || gzipped_file_mode)
+	      if (compressed_file_mode || gzipped_file_mode
+		  || bzipped_file_mode)
 		pclose (input);
 	      else
 		fclose (input);
@@ -1303,17 +1415,17 @@ else\n",
 	  if (split_file_mode
 #if MSDOS
 	      /* 1 extra for CR.  */
-	      && (remaining_size -= (int) strlen (buffer) + 1) < 0
+	      && (remaining_size -= strlen (buffer) + 1) < 0
 #else
-	      && (remaining_size -= (int) strlen (buffer)) < 0
+	      && (remaining_size -= strlen (buffer)) < 0
 #endif
 	      )
 	    {
 
 	      /* Change to another file.  */
 
-	      DEBUG_PRINT (_("Newfile, remaining %ld, "), remaining_size);
-	      DEBUG_PRINT (_("Limit still %d\n"), file_size_limit);
+	      DEBUG_PRINT (_("New file, remaining %s, "), remaining_size);
+	      DEBUG_PRINT (_("Limit still %s\n"), file_size_limit);
 
 	      fprintf (output, "%s\n", here_delimiter);
 
@@ -1355,7 +1467,7 @@ exit 0\n",
 
 		  /* Rewrite the info lines on the first header.  */
 
-		  fseek (output, archive_type_position, 0);
+		  fseeko (output, archive_type_position, SEEK_SET);
 		  fprintf (output, "%-75s\n%-75s\n",
 			   "\
 # This is part 1 of a multipart archive.",
@@ -1410,17 +1522,19 @@ fi\n",
 		       N_("next!"));
 
 	      if (check_existing_mode)
-		if (quiet_unshar_mode)
-		  fprintf (output, "\
+		{
+		  if (quiet_unshar_mode)
+		    fprintf (output, "\
 if test -f _sh%05d/new; then\n",
-			   sharpid);
-		else
-		  fprintf (output, "\
+			     sharpid);
+		  else
+		    fprintf (output, "\
 if test ! -f _sh%05d/new; then\n\
   $echo 'x -' '%s' '%s'\n\
 else\n",
-			   sharpid,
-			   N_("STILL SKIPPING"), restore_name);
+			     sharpid,
+			     N_("STILL SKIPPING"), restore_name);
+		}
 
 	      if (!quiet_mode)
 		fprintf (stderr, _("Starting file %s\n"), output_filename);
@@ -1435,7 +1549,7 @@ else\n",
 		fprintf (output, "_sh%05d/uue &&\n", sharpid);
 	      else
 		fprintf (output, "%s &&\n", restore_name);
-	      remaining_size = file_size_limit * 1024L;
+	      remaining_size = file_size_limit;
 	      split_flag = 1;
 	    }
 	}
@@ -1488,6 +1602,17 @@ else\n",
   gzip -d < _sh%05d/gzi > '%s' &&\n",
 		   sharpid, restore_name);
 	}
+      else if (bzipped_file_mode)
+	{
+	  if (!quiet_unshar_mode)
+	    fprintf (output, "\
+  $echo '%s' '%s' &&\n",
+		     N_("bunzipping file"), restore_name);
+
+	  fprintf (output, "\
+  bzip2 -d < _sh%05d/bzi > '%s' &&\n",
+		   sharpid, restore_name);
+	}
     }
 
   if (timestamp_mode)
@@ -1541,19 +1666,20 @@ else\n",
 	  did_md5 = 1;
 
 	  fprintf (output, "\
-  if ( %s --help 2>&1 | grep 'sage: md5sum \\[' ) >/dev/null 2>&1 \\\n\
-  && ( %s --version 2>&1 | grep -v 'textutils 1.12' ) >/dev/null; then\n\
+  if ( %s --help </dev/null 2>&1 | grep 'sage: md5sum \\[' ) >/dev/null 2>&1 \\\n\
+  && ( %s --version </dev/null 2>&1 | grep -v 'textutils 1.12' ) >/dev/null; then\n\
     %s -c << %s >/dev/null 2>&1 \\\n\
     || $echo '%s:' '%s'\n",
 		   MD5SUM_COMMAND, MD5SUM_COMMAND, MD5SUM_COMMAND,
-		   DEFAULT_HERE_DELIMITER, local_name, N_("MD5 check failed"));
+		   DEFAULT_HERE_DELIMITER, restore_name,
+		   N_("MD5 check failed"));
 
 	  for (cnt = 0; cnt < 16; ++cnt)
 	    fprintf (output, "%02x", md5buffer[cnt]);
 
 	  fprintf (output, " %c%s\n\
 %s\n",
-		   ' ', local_name, DEFAULT_HERE_DELIMITER);
+		   ' ', restore_name, DEFAULT_HERE_DELIMITER);
 	  /* This  ^^^ space is not necessarily a parameter now.  But it
 	     is a flag for binary/text mode and will perhaps be used later.  */
 	}
@@ -1571,7 +1697,7 @@ else\n",
 	  sprintf (command, "%s '%s'", CHARACTER_COUNT_COMMAND, local_name);
 	  if (pfp = popen (command, "r"), pfp)
 	    {
-	      char wc[BUFSIZ], tempform[50];
+	      char wc[BUFSIZ];
 	      const char *prefix = "";
 
 	      if (did_md5)
@@ -1580,8 +1706,27 @@ else\n",
 		  prefix = "  ";
 		}
 
-	      sprintf (tempform, "%%%ds", BUFSIZ - 1);
-	      fscanf (pfp, tempform, wc);
+              {
+                char* pz = wc;
+                for (;;)
+                  {
+		    int ch = (unsigned int)fgetc (pfp);
+		    if (! isspace (ch))
+		      {
+			*(pz++) = ch;
+			break;
+		      }
+                  }
+                if (pz[-1] != NUL) do
+                  {
+                    int ch = (unsigned int)fgetc (pfp);
+                    if ((ch < 0) || isspace (ch))
+                      break;
+                    *pz = ch;
+                  } while (++pz < wc+BUFSIZ-1);
+                *pz = NUL;
+              }
+
 	      fprintf (output, "\
 %s  shar_count=\"`%s '%s'`\"\n\
 %s  test %s -eq \"$shar_count\" ||\n\
@@ -1611,7 +1756,8 @@ fi\n",
 
 /*-----------------------------------------------------------------------.
 | Set file mode, accepting a parameter 'M' for mixed uuencoded mode, 'B' |
-| for uuencoded mode, 'z' for gzipped mode or 'Z' for compressed mode.	 |
+| for uuencoded mode, 'z' for gzipped mode, 'j' for bzipped mode, or 'Z' |
+| for compressed mode.							 |
 | Any other value yields text mode.					 |
 `-----------------------------------------------------------------------*/
 
@@ -1619,12 +1765,17 @@ static void
 set_file_mode (mode)
      int mode;
 {
+  if (mode == 'B' && uuencoded_file_mode)
+    /* Selecting uuencode mode should not contradict compression.  */
+    return;
+
   mixed_uuencoded_file_mode = mode == 'M';
   uuencoded_file_mode = mode == 'B';
   gzipped_file_mode = mode == 'z';
+  bzipped_file_mode = mode == 'j';
   compressed_file_mode = mode == 'Z';
 
-  if (gzipped_file_mode || compressed_file_mode)
+  if (gzipped_file_mode || bzipped_file_mode || compressed_file_mode)
     uuencoded_file_mode = 1;
 }
 
@@ -1635,13 +1786,12 @@ set_file_mode (mode)
 static void
 open_output ()
 {
-  size_t l;
-  l = strlen(output_base_name) + 128;
-  if (output_filename)
-    free(output_filename);
-  output_filename = xmalloc(l);
-  snprintf(output_filename, l, output_base_name, ++part_number);
+  if (output_filename == NULL)
+    error (EXIT_FAILURE, ENXIO, _("allocating output file name"));
+
+  sprintf(output_filename, output_base_name, ++part_number);
   output = fopen (output_filename, "w");
+
   if (!output)
     error (EXIT_FAILURE, errno, _("Opening `%s'"), output_filename);
 }
@@ -1679,11 +1829,24 @@ Mandatory arguments to long options are mandatory for short options too.\n"),
 Giving feedback:\n\
       --help              display this help and exit\n\
       --version           output version information and exit\n\
-  -q, --quiet, --silent   do not output verbose messages locally\n\
+  -q, --quiet, --silent   do not output verbose messages locally\n"),
+	     stdout);
+#if HAVE_COMPRESS
+      fputs (_("\
 \n\
 Selecting files:\n\
   -p, --intermix-type     allow -[BTzZ] in file lists to change mode\n\
-  -S, --stdin-file-list   read file list from standard input\n\
+  -S, --stdin-file-list   read file list from standard input\n"),
+	     stdout);
+#else
+      fputs (_("\
+\n\
+Selecting files:\n\
+  -p, --intermix-type     allow -[BTz] in file lists to change mode\n\
+  -S, --stdin-file-list   read file list from standard input\n"),
+	     stdout);
+#endif
+      fputs (_("\
 \n\
 Splitting output:\n\
   -o, --output-prefix=PREFIX    output to file PREFIX.01 through PREFIX.NN\n\
@@ -1704,9 +1867,14 @@ Selecting how files are stocked:\n\
   -B, --uuencode               treat all files as binary, use uuencode\n\
   -z, --gzip                   gzip and uuencode all files\n\
   -g, --level-for-gzip=LEVEL   pass -LEVEL (default 9) to gzip\n\
+  -j, --bzip2                  bzip2 and uuencode all files\n"),
+	     stdout);
+#if HAVE_COMPRESS
+      fputs (_("\
   -Z, --compress               compress and uuencode all files\n\
   -b, --bits-per-code=BITS     pass -bBITS (default 12) to compress\n"),
 	     stdout);
+#endif
       fputs (_("\
 \n\
 Protecting against transmission:\n\
@@ -1727,9 +1895,19 @@ Producing different kinds of shars:\n\
 	     stdout);
       fputs (_("\
 \n\
-Option -o is required with -l or -L, option -n is required with -a.\n\
+Option -o is required with -l or -L, option -n is required with -a.\n"),
+	     stdout);
+#if HAVE_COMPRESS
+      fputs (_("\
 Option -g implies -z, option -b implies -Z.\n"),
 	     stdout);
+#else
+      fputs (_("\
+Option -g implies -z.\n"),
+	     stdout);
+#endif
+      /* TRANSLATORS: add the contact address for your translation team! */
+      printf (_("Report bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
   exit (status);
 }
@@ -1743,6 +1921,7 @@ static const struct option long_options[] =
   {"archive-name", required_argument, NULL, 'n'},
   {"basename", no_argument, NULL, 'f'},
   {"bits-per-code", required_argument, NULL, 'b'},
+  {"bzip2", no_argument, NULL, 'j'},
   {"compress", no_argument, NULL, 'Z'},
   {"cut-mark", no_argument, NULL, 'c'},
   {"force-prefix", no_argument, NULL, 'F'},
@@ -1777,38 +1956,73 @@ static const struct option long_options[] =
   { NULL, 0, NULL, 0 },
 };
 
+/* Limit file sizes to LIMIT KiB.  */
 
-char *parse_output_base_name(char *arg)
+static void
+set_file_size_limit (char const *limit)
+{
+  char *numend;
+  intmax_t lim = strtoimax (limit, &numend, 10);
+  if (*numend || ! (0 < lim && lim <= TYPE_MAXIMUM (off_t) / 1024))
+    error (EXIT_FAILURE, 0, _("invalid file size limit `%s'"), limit);
+  lim *= 1024;
+  if (! file_size_limit || lim < file_size_limit)
+    file_size_limit = lim;
+}
+
+
+char *
+parse_output_base_name(char *arg)
 {
   int c;
   int hadarg = 0;
   char *fmt, *p;
+  int base_name_len = 128;
 
   for (p = arg ; (c = *p++) != 0; )
     {
+      base_name_len++;
       if (c != '%')
 	continue;
       c = *p++;
       if (c == '%')
 	continue;
       if (hadarg)
-	return 0;
+	return NULL;
       while (c != 0 && strchr("#0+- 'I", c) != 0)
 	c = *p++;
-      while (c != 0 && c >= '0' && c <= '9')
-	c = *p++;
+      if (c == 0)
+	return NULL;
+      if (c >= '0' && c <= '9')
+	{
+	  long v;
+	  errno = 0;
+	  v = strtol(p-1, &fmt, 10);
+	  if ((v == 0) || (v > 16) || (errno != 0))
+	    {
+	      fprintf(stderr, _("invalid format (count field too wide): '%s'\n"),
+		      arg);
+	      return NULL;
+	    }
+	  p = fmt;
+	  c = *p++;
+	  base_name_len += v;
+	}
       if (c == '.')
-	c = *p++;
-      while (c != 0 && c >= '0' && c <= '9')
-	c = *p++;
+	{
+	  c = *p++;
+	  while (c != 0 && c >= '0' && c <= '9')
+	    c = *p++;
+	}
       if (c == 0 || strchr("diouxX", c) == 0)
-	return 0;
+	return NULL;
       hadarg = 1;
     }
   fmt = xmalloc(strlen(arg) + (hadarg ? 1 : 6));
   strcpy(fmt, arg);
   if (!hadarg)
     strcat(fmt, ".%02d");
+  output_filename = xmalloc(base_name_len);
   return fmt;
 }
 
@@ -1835,12 +2049,12 @@ main (argc, argv)
   textdomain (PACKAGE);
 
   while (optchar = getopt_long (argc, argv,
-				"+$BCDFL:MPQSTVXZab:cd:fg:hl:mn:o:pqs:wxz",
+				"+$BCDFL:MPQSTVXZab:cd:fg:hjl:mn:o:pqs:wxz",
 				long_options, NULL),
 	 optchar != EOF)
     switch (optchar)
       {
-      case '\0':
+      case NUL:
 	break;
 
       case '$':
@@ -1864,11 +2078,10 @@ main (argc, argv)
 	break;
 
       case 'L':
-	if (file_size_limit = atoi (optarg), file_size_limit > 1)
-	  file_size_limit--;
-	split_file_mode = file_size_limit != 0;
+	set_file_size_limit (optarg);
+	split_file_mode = 1;
 	inhibit_piping_mode = 1;
-	DEBUG_PRINT (_("Hard limit %dk\n"), file_size_limit);
+	DEBUG_PRINT (_("Hard limit %s\n"), file_size_limit);
 	break;
 
       case 'M':
@@ -1906,6 +2119,10 @@ main (argc, argv)
 
       case 'C':
       case 'Z':
+#ifndef HAVE_COMPRESS
+	error (EXIT_FAILURE, 0, _("\
+This system doesn't support -Z ('compress'), use -z instead"));
+#endif
 	if (optchar == 'C')
 	  error (0, 0, _("-C is being deprecated, use -Z instead"));
 	set_file_mode ('Z');
@@ -1931,11 +2148,14 @@ main (argc, argv)
 	usage (EXIT_SUCCESS);
 	break;
 
+      case 'j':
+	set_file_mode ('j');
+	break;
+
       case 'l':
-	if (file_size_limit = atoi (optarg), file_size_limit > 1)
-	  file_size_limit--;
+	set_file_size_limit (optarg);
 	split_file_mode = 0;
-	DEBUG_PRINT (_("Soft limit %dk\n"), file_size_limit);
+	DEBUG_PRINT (_("Soft limit %s\n"), file_size_limit);
 	break;
 
       case 'm':
@@ -1947,8 +2167,6 @@ main (argc, argv)
 	break;
 
       case 'o':
-	if (output_base_name)
-	  free (output_base_name);
         output_base_name = parse_output_base_name(optarg);
         if (!output_base_name)
 	  {
@@ -1997,7 +2215,13 @@ main (argc, argv)
 
   if (show_version)
     {
-      printf ("%s - GNU %s %s\n", program_name, PACKAGE, VERSION);
+      printf ("%s (GNU %s) %s\n", basename (program_name), PACKAGE, VERSION);
+      /* xgettext: no-wrap */
+      printf (_("Copyright (C) %s Free Software Foundation, Inc.\n\
+This is free software; see the source for copying conditions.  There is NO\n\
+warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
+"),
+	      "1994, 1995, 1996");
       exit (EXIT_SUCCESS);
     }
 
@@ -2012,8 +2236,7 @@ main (argc, argv)
     || (defined HAVE_GETTEXT && !defined __USE_GNU_GETTEXT)
       exit (EXIT_FAILURE);
 #else
-      extern const char _nl_default_dirname[]; /* Defined in dcgettext.c  */
-      puts (_nl_default_dirname);
+      puts (LOCALEDIR);
       exit (EXIT_SUCCESS);
 #endif
     }
@@ -2050,6 +2273,7 @@ main (argc, argv)
       if (mixed_uuencoded_file_mode
 	  || uuencoded_file_mode
 	  || gzipped_file_mode
+	  || bzipped_file_mode
 	  || compressed_file_mode
 	  || intermixed_parameter_mode)
 	error (0, 0, _("WARNING: Non-text storage options overridden"));
@@ -2094,7 +2318,7 @@ main (argc, argv)
 	  if (argc == max_argc)
 	    list = (char **) xrealloc (list,
 				       (max_argc *= 2) * sizeof (char *));
-	  if (stdin_buf[0] != '\0')
+	  if (stdin_buf[0] != NUL)
 	    stdin_buf[strlen (stdin_buf) - 1] = 0;
 	  list[argc] = xstrdup (stdin_buf);
 	  ++argc;
@@ -2135,37 +2359,37 @@ main (argc, argv)
 	error (0, 0, _("PLEASE avoid -X shars on Usenet or public networks"));
 
       fputs ("\
-shar_wish=\n",
-	     output);
+shar_wish=\n", output);
     }
 
-  first_file_position = ftell (output);
+  first_file_position = ftello (output);
 
   /* Process positional parameters and files.  */
 
   for (; optind < argc; optind++)
-    if (intermixed_parameter_mode)
-      if (strcmp (argv[optind], "-B") == 0)
-	set_file_mode ('B');
-      else if (strcmp (argv[optind], "-T") == 0)
-	set_file_mode ('T');
-      else if (strcmp (argv[optind], "-M") == 0)
-	set_file_mode ('M');
-      else if (strcmp (argv[optind], "-z") == 0)
-	set_file_mode ('z');
-      else if (strcmp (argv[optind], "-Z") == 0
-	       || strcmp (argv[optind], "-C") == 0)
-	set_file_mode ('Z');
-      else
-	{
-	  if (walktree (shar, argv[optind]))
-	    status = EXIT_FAILURE;
-	}
-    else
-      {
-	if (walktree (shar, argv[optind]))
-	  status = EXIT_FAILURE;
-      }
+    {
+      char* arg = argv[optind];
+
+      if (intermixed_parameter_mode && (arg[0] == '-') && (arg[2] == NUL))
+        {
+          switch (arg[1]) {
+          case 'B':
+          case 'T':
+          case 'M':
+          case 'z':
+          case 'Z':
+            set_file_mode (arg[1]);
+            continue;
+
+          case 'C':
+            set_file_mode ('Z');
+            continue;
+          }
+        }
+
+      if (walktree (shar, arg))
+        status = EXIT_FAILURE;
+    }
 
   /* Delete the sequence file, if any.  */
 
@@ -2179,9 +2403,15 @@ $echo '%s'\n",
     }
 
   fprintf (output, "\
-rm -fr _sh%05d\n\
+$echo $shar_n 'x -' '%s' \"\\`_sh%05d': \" $shar_c\n\
+if rm -fr _sh%05d; then\n\
+  $echo '%s'\n\
+else\n\
+  $echo '%s'\n\
+fi\n\
 exit 0\n",
-	 sharpid);
+	   N_("lock directory"), sharpid, sharpid, N_("removed"),
+	   N_("failed to remove"));
 
   exit (status);
 }
